from typing import Optional, List


from rhazes.dependency import DependencyResolver
from rhazes.protocol import BeanProtocol
from rhazes.scanner import ModuleScanner, class_scanner
from rhazes.utils import LazyObject


class ApplicationContext:
    _initialized = False
    _builder_registry = {}

    @classmethod
    def _initialize_beans(cls, packages_to_scan: List[str]):
        beans = set()
        modules = ModuleScanner(packages_to_scan).scan()
        for module in modules:
            for bean in class_scanner(
                module, lambda clz: issubclass(clz, (BeanProtocol,))
            ):
                beans.add(bean)

        for clazz, obj in DependencyResolver(beans).resolve().items():
            cls.register_bean(clazz, obj)

    @classmethod
    def initialize(cls, packages_to_scan: List[str]):
        if packages_to_scan is not None:
            cls.packages_to_scan = packages_to_scan

        if cls._initialized:
            return
        cls._initialize_beans(packages_to_scan)
        cls._initialized = True

    @classmethod
    def register_bean(cls, clazz, builder, override=False):
        if clazz not in cls._builder_registry or override:
            cls._builder_registry[clazz] = builder

    @classmethod
    def get_bean(cls, of: type) -> Optional:
        builder = cls._builder_registry.get(of)
        if builder is None:
            return None
        return builder(cls)

    @classmethod
    def get_lazy_bean(cls, of: type) -> Optional:
        return LazyObject(lambda: cls.get_bean(of))



import operator


class LazyObject:

    _wrapped = None
    _is_init = False

    def __init__(self, factory):
        # Assign using __dict__ to avoid the setattr method.
        self.__dict__["_factory"] = factory

    def _setup(self):
        self._wrapped = self._factory()
        self._is_init = True

    def new_method_proxy(func):
        """
        Util function to help us route functions
        to the nested object.
        """

        def inner(self, *args):
            if not self._is_init:
                self._setup()
            return func(self._wrapped, *args)

        return inner

    def __setattr__(self, name, value):
        # These are special names that are on the LazyObject.
        # every other attribute should be on the wrapped object.
        if name in {"_is_init", "_wrapped"}:
            self.__dict__[name] = value
        else:
            if not self._is_init:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        if name == "_wrapped":
            raise TypeError("can't delete _wrapped.")
        if not self._is_init:
            self._setup()
        delattr(self._wrapped, name)

    __getattr__ = new_method_proxy(getattr)
    __bytes__ = new_method_proxy(bytes)
    __str__ = new_method_proxy(str)
    __bool__ = new_method_proxy(bool)
    __dir__ = new_method_proxy(dir)
    __hash__ = new_method_proxy(hash)
    __class__ = property(new_method_proxy(operator.attrgetter("__class__")))
    __eq__ = new_method_proxy(operator.eq)
    __lt__ = new_method_proxy(operator.lt)
    __gt__ = new_method_proxy(operator.gt)
    __ne__ = new_method_proxy(operator.ne)
    __hash__ = new_method_proxy(hash)
    __getitem__ = new_method_proxy(operator.getitem)
    __setitem__ = new_method_proxy(operator.setitem)
    __delitem__ = new_method_proxy(operator.delitem)
    __iter__ = new_method_proxy(iter)
    __len__ = new_method_proxy(len)
    __contains__ = new_method_proxy(operator.contains)
